#include<avr/io.h>
#include<avr/interrupt.h>

#define TASK_ANZAHL 2

void (*taskfunktion[TASK_ANZAHL])(uint8_t*,uint8_t*);
volatile uint8_t  DELAY[TASK_ANZAHL];
volatile uint8_t STATUS[TASK_ANZAHL];

volatile uint32_t ZEIT = 0;

SIGNAL(SIG_OUTPUT_COMPARE2)
{
    int i;
    TCNT2=0;
    for(i=0;i<TASK_ANZAHL;i++)
        if(DELAY[i]>0)
            DELAY[i]--;	
}

void uhr(uint8_t* DELAY,uint8_t* STATUS)
{
    *DELAY = 200;  //Uhr: zählt im Sekundentakt.
    *STATUS = 0;
    ZEIT++;
    PORTC = ZEIT;	
}

void ton(uint8_t* DELAY,uint8_t* STATUS)
{
    *DELAY = 2;  //Ton von 50Hz.
    switch(*STATUS)
    {
        case 0:
            *STATUS = 1;
            PORTA |= 0b00000100;
        break;
        case 1:
            *STATUS = 0;
            PORTA &= 0b11111011; 
        break;	 	
    }
}

int main()
{
    int i;	
    //Timer2 in CTC-Mode ohne externe Ausgänge:
    TCCR2 = (0<<FOC2) | (0<<WGM20) | (0<<COM21) | (1<<COM20) | (1<<WGM21)
          | (1<<CS22) | (1<<CS21) | (0<<CS20);	
    //CS22 CS21 CS20  Teilung
    //001 /1          
    //010 /8
    //011 /32
    //100 /64
    //101 /128
    //110 /256
    //111 /1024          
    // fctc = fclock/(2*N*(1+OCR2))      
    //N=256, OCR2=17 => 1000Hz, 500Hz bei OC2.  
    //N=256, OCR2=170 => 100Hz, 50Hz bei OC2.  
    //Toggle on Compare-Match => Halbe Frequenz abrufbar auf Pin 21, OC2
    //Start mit niedriger gut wahrnehmbaerer Frequenz:
    //N=1024, OCR2=249 => 18Hz, 9Hz an OC2.     
    //N=1024, OCR2=149 => 30Hz, 15Hz an OC2.
    
    //Tatsächlich:    
    // fctc = fclock/(N*(1+OCR2))
    //N=256, OCR2=17 => 2000Hz, 1000Hz bei OC2.  
    //N=256, OCR2=179 => 200Hz, 100Hz bei OC2.  
          
    
    OCR2=179;    
    
    //Timermaske: Interrupt wenn Compare-Match bei Timer2:
    TIMSK |= (1<<OCIE2);
    
    //Interrupts aktivieren:
    sei();

    //PD7 als Ausgang, um die halbe Interruptfrequenz im CTC-Modus dort messbar zu machen.
    DDRD|=0b10000000;
    
    //PA2 als Ausgang    
    DDRA|=0b00000100;
    
    //Test: 1Hz Blinken auf PORTC:
    DDRC=255;
    PORTC=0;

    taskfunktion[0] = ton;
    taskfunktion[1] = uhr;
            
    
    while(1)            
    {
        for(i=0;i<TASK_ANZAHL;i++)
        {    	
            if(DELAY[i]==0)
            {
                (taskfunktion[i])(&DELAY[i],&STATUS[i]);
            }
        }    	
    }
}


