#include <avr/io.h>
#include <avr/interrupt.h>

#define TASK_ANZAHL 3
#define TAKTFREQUENZ 9216000
#define BAUDRATE 115200
#define YIELD_DELAY 255

void (*taskfunktion[TASK_ANZAHL])(uint8_t*,uint8_t*);
volatile uint8_t  DELAY[TASK_ANZAHL];
volatile uint8_t STATUS[TASK_ANZAHL];

volatile uint32_t ZEIT = 0;

SIGNAL(SIG_OUTPUT_COMPARE2)
{
	int i;
	TCNT2=0;
	for(i=0;i<TASK_ANZAHL;i++)
	if((DELAY[i]>0)  &&  (DELAY[i]<YIELD_DELAY))
    DELAY[i]--;
}

void uhr(uint8_t* DELAY,uint8_t* STATUS)
{
	*DELAY = 200;  //Uhr: zhlt im Sekundentakt.
	*STATUS = 0;
	ZEIT++;
	PORTC = ZEIT;
}

void ton(uint8_t* DELAY,uint8_t* STATUS)
{
	*DELAY = 2;  //Ton von 50Hz.
	switch(*STATUS)
	{
		case 0:
		*STATUS = 1;
		PORTA |= 0b00000100;
		break;
		case 1:
		*STATUS = 0;
		PORTA &= 0b11111011;
		break;
	}
}

void RS232(uint8_t* DELAY,uint8_t* STATUS)
{
    *DELAY = 255;

	switch(*STATUS)
	{
	    case 0:
        if( (UCSRA & (1<<UDRE)) );     //Warten bis der Uebertragungspuffer leer ist
        {
		    UDR = 'H';                 //Daten in den Puffer schreiben und bertragen
	        *STATUS = 1;
		}
		break;
		case 1:
        if( (UCSRA & (1<<UDRE)) );     //Warten bis der Uebertragungspuffer leer ist
        {
		    UDR = 'A';                 //Daten in den Puffer schreiben und bertragen
	        *STATUS = 2;
		}
		break;
		case 2:
        if( (UCSRA & (1<<UDRE)) );     //Warten bis der Uebertragungspuffer leer ist
        {
		    UDR = 'L';                 //Daten in den Puffer schreiben und bertragen
	        *STATUS = 3;
		}
		break;
		case 3:
        if( (UCSRA & (1<<UDRE)) );     //Warten bis der Uebertragungspuffer leer ist
        {
		    UDR = 'L';                 //Daten in den Puffer schreiben und bertragen
	        *STATUS = 4;
		}
		break;
		case 4:
        if( (UCSRA & (1<<UDRE)) );     //Warten bis der Uebertragungspuffer leer ist
        {
		    UDR = 'O';                 //Daten in den Puffer schreiben und bertragen
	        *STATUS = 5;
		}
		break;
		case 5:
        if( (UCSRA & (1<<UDRE)) );     //Warten bis der Uebertragungspuffer leer ist
        {
		    UDR = ' ';                 //Daten in den Puffer schreiben und bertragen
	        *STATUS = 0;
		}
		break;
	}
}

int main()
{
	int i;
	int index=0;
	//Timer2 in CTC-Mode ohne externe Ausgänge:
	TCCR2 = (0<<FOC2) | (0<<WGM20) | (0<<COM21) | (1<<COM20) | (1<<WGM21)
	| (1<<CS22) | (1<<CS21) | (0<<CS20);
	//CS22 CS21 CS20  Teilung
	//001 /1
	//010 /8
	//011 /32
	//100 /64
	//101 /128
	//110 /256
	//111 /1024
	// fctc = fclock/(2*N*(1+OCR2))
	//N=256, OCR2=17 => 1000Hz, 500Hz bei OC2.
	//N=256, OCR2=170 => 100Hz, 50Hz bei OC2.
	//Toggle on Compare-Match => Halbe Frequenz abrufbar auf Pin 21, OC2
	//Start mit niedriger gut wahrnehmbaerer Frequenz:
	//N=1024, OCR2=249 => 18Hz, 9Hz an OC2.
	//N=1024, OCR2=149 => 30Hz, 15Hz an OC2.
	
	//Tatsächlich:
	// fctc = fclock/(N*(1+OCR2))
	//N=256, OCR2=17 => 2000Hz, 1000Hz bei OC2.
	//N=256, OCR2=179 => 200Hz, 100Hz bei OC2.
	
	
	OCR2=179;
	
	//Timermaske: Interrupt wenn Compare-Match bei Timer2:
	TIMSK |= (1<<OCIE2);
	
	//Interrupts aktivieren:
	sei();

	//PD7 als Ausgang, um die halbe Interruptfrequenz im CTC-Modus dort messbar zu machen.
	DDRD|=0b10000000;
	
	//PA2 als Ausgang
	DDRA|=0b00000100;
	
	//Test: 1Hz Blinken auf PORTC:
	DDRC=255;
	PORTC=0;

	taskfunktion[0] = ton;
	taskfunktion[1] = uhr;
	taskfunktion[2] = RS232;
	
    //Merken des in UBRR zu speichernden Wertes.
    unsigned int baudregister = TAKTFREQUENZ/8/BAUDRATE-1;

    //setzen der Baudrate
    UBRRH = (unsigned char) (baudregister>>8); //Setzen des HIGH-Bytes des Baudraten-Registers
    UBRRL = (unsigned char)  baudregister;     //Setzen des LOW -Bytes des Baudraten-Registers
	
    //Einschalten des Senders und des Empfngers
    UCSRB = (1<<TXEN) | (1<<RXEN);
		        
    //Setzen des Nachrichtenformats: 8 Datenbits, 1 Stopbits
    UCSRC = (1<<URSEL)|(1<<UCSZ0)|(1<<UCSZ1);
    UCSRA = (1<<U2X);	


	while(1)
	{

		for(i=0;i<TASK_ANZAHL;i++)
		{
			if(DELAY[i]==0)
			{
				(taskfunktion[i])(&DELAY[i],&STATUS[i]);
			}
		}
		if (DELAY[index]==YIELD_DELAY)
		   (taskfunktion[index])(&DELAY[index],&STATUS[index]);
        index++;
		index%=TASK_ANZAHL;
	}
}